# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import shutil
from os import listdir
from argparse import ArgumentParser
from os.path import join, abspath, expanduser, basename

from git import Git, GitCommandError
from msm import SkillEntry

from msk.actions.create import CreateAction
from msk.console_action import ConsoleAction
from msk.exceptions import (
    MskException,
    NoGitRepository,
    UnrelatedGithubHistory,
    GithubRepoExists,
)
from msk.lazy import Lazy
from msk.repo_action import SkillData
from msk.util import (
    skills_kit_footer,
    create_or_edit_pr,
    ask_yes_no,
    skill_repo_name,
    read_file,
)

body_template = (
    """
# Info

This PR adds the new skill, [{skill_name}]({skill_url}), to the skills repo.

# Description

{description}

"""
    + skills_kit_footer
)


def exists_in_remote(git, repo_path):
    """Checks if the file at repo_path exists in remote HEAD."""
    try:
        remote_info = git.remote("show", "origin").split("\n")
        heads = [e.split(":")[1].strip() for e in remote_info if "HEAD" in e]
        remote_file = "origin/{}:{}".format(heads[0], repo_path)
        git.cat_file(remote_file, e=True)
    except GitCommandError:
        return False
    else:
        return True


class UploadAction(ConsoleAction):
    def __init__(self, args):
        folder = abspath(expanduser(args.skill_folder))
        self.entry = SkillEntry.from_folder(folder)
        skills_dir = abspath(expanduser(self.msm.skills_dir))
        if join(skills_dir, basename(folder)) != folder:
            raise MskException(
                f"Skill folder, {args.skill_folder}, not directly within "
                f"skills directory, {self.msm.skills_dir}."
            )
        self.skill_dir = folder

    git = Lazy(lambda s: Git(s.entry.path))  # type: Git

    @staticmethod
    def register(parser: ArgumentParser):
        pass  # Implemented in SubmitAction

    def check_valid(self):
        """Check the skill contains all required files before uploading."""
        results = []
        if not (
            exists_in_remote(self.git, "LICENSE.md")
            or exists_in_remote(self.git, "LICENSE")
            or exists_in_remote(self.git, "LICENSE.txt")
        ):
            print(
                "To have your Skill available for installation through the "
                "Skills Marketplace, a license is required.\n"
                "Please select one and add it to the skill as "
                "`LICENSE.md.`\n"
                "See https://opensource.org/licenses for information on "
                "open source license options."
            )
            results.append(False)
        else:
            results.append(True)

        if not exists_in_remote(self.git, "README.md"):
            print(
                "For inclusion in the Mycroft Marketplace a README.md file "
                "is required. please add the file and retry."
            )
            results.append(False)
        else:
            results.append(True)

        with open(join(self.skill_dir, "README.md")) as f:
            readme = f.read()
        if "# About" not in readme and "# Description" not in readme:
            print("README is missing About Section needed by the Marketplace")
            results.append(False)
        else:
            results.append(True)

        if "# Category" not in readme:
            print(
                "README is missing Category section needed by the "
                "Marketplace"
            )
            results.append(False)
        else:
            results.append(True)
        return all(results)

    def perform(self):
        print("Uploading a new skill to the skill repo...")

        for i in listdir(self.entry.path):
            if i.lower() == "readme.md" and i != "README.md":
                shutil.move(
                    join(self.entry.path, i),
                    join(self.entry.path, "README.md"),
                )

        creator = CreateAction(None, self.entry.name.replace("-skill", ""))
        creator.path = self.entry.path
        creator.initialize_template({".git", ".gitignore", "README.md"})
        self.git.add("README.md")
        creator.commit_changes()

        try:
            skill_repo = creator.create_github_repo(
                lambda: input("Repo name:")
            )
        except GithubRepoExists:
            try:
                print("A repository with that name already exists")
                skill_repo = creator.link_github_repo(
                    lambda: input("Remote repo name:")
                )
            except UnrelatedGithubHistory:
                print("Repository history does not seem to be related")
                skill_repo = creator.force_push(
                    lambda: input("Confirm repo name:")
                )
        if skill_repo:
            self.entry.url = skill_repo.html_url
            self.entry.author = self.user.login
        else:
            if not self.entry.url:
                raise NoGitRepository
            skill_repo = self.github.get_repo(skill_repo_name(self.entry.url))

        if not skill_repo.permissions.push:
            print(
                "Warning: You do not have write permissions to the provided "
                "skill repo."
            )
            if ask_yes_no("Create a fork and use that instead? (Y/n)", True):
                skill_repo = self.user.create_fork(skill_repo)
                print("Created fork:", skill_repo.html_url)
                self.git.remote("rename", "origin", "upstream")
                self.git.remote("add", "origin", skill_repo.html_url)

        # verify that the required files exists in origin and contain the
        # required content.
        if not self.check_valid():
            print("Please add the missing information and rerun the command.")
            return

        self.entry.name = input(
            "Enter a unique skill name (ie. npr-news or grocery-list): "
        )

        readme_file = {i.lower(): i for i in os.listdir(self.entry.path)}[
            "readme.md"
        ]
        readme = read_file(self.entry.path, readme_file)

        last_section = None
        sections = {last_section: ""}
        for line in readme.split("\n"):
            line = line.strip()
            if line.startswith("#"):
                last_section = line.strip("# ").lower()
                sections[last_section] = ""
            else:
                sections[last_section] += "\n" + line
        del sections[None]

        if "about" in sections:
            description = sections["about"]
        elif "description" in sections:
            description = sections["description"]

        branch = SkillData(self.entry).add_to_repo()
        self.repo.push_to_fork(branch)

        pull = create_or_edit_pr(
            title=f"Add {self.entry.name}",
            body=body_template.format(
                description=description,
                skill_name=self.entry.name,
                skill_url=skill_repo.html_url,
            ),
            user=self.user,
            branch=branch,
            skills_repo=self.repo.hub,
            repo_branch=self.branch,
        )

        print("Created pull request: ", pull.html_url)



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys

from argparse import ArgumentParser
from msm import MycroftSkillsManager, SkillRepo

from msk.actions.create import CreateAction
from msk.actions.create_test import CreateTestAction
from msk.actions.submit import SubmitAction
from msk.exceptions import MskException
from msk.global_context import GlobalContext
from msk.util import ensure_git_user

action_names = {
    SubmitAction: ["submit", "update", "upgrade", "upload"],
    CreateAction: ["create"],
    CreateTestAction: ["create-test"],
}


def main():
    parser = ArgumentParser()
    parser.add_argument("-l", "--lang", default="en-us")
    parser.add_argument(
        "-u", "--repo-url", help="Url of GitHub repo to upload skills to"
    )
    parser.add_argument(
        "-b", "--repo-branch", help="Branch of skills repo to upload to"
    )
    parser.add_argument(
        "-s", "--skills-dir", help="Directory to look for skills in"
    )
    parser.add_argument(
        "-c", "--repo-cache", help="Location to store local skills repo clone"
    )

    subparsers = parser.add_subparsers(dest="action")
    subparsers.required = True
    action_to_cls = {}
    for cls, names in action_names.items():
        cls.register(subparsers.add_parser(names[0], aliases=names[1:]))
        action_to_cls.update({name: cls for name in names})

    args = parser.parse_args(sys.argv[1:])

    ensure_git_user()
    context = GlobalContext()
    context.lang = args.lang
    context.msm = MycroftSkillsManager(
        skills_dir=args.skills_dir,
        repo=SkillRepo(
            url=args.repo_url, branch=args.repo_branch, path=args.repo_cache
        ),
    )
    context.branch = context.msm.repo.branch

    try:
        return action_to_cls[args.action](args).perform()
    except MskException as e:
        print("{}: {}".format(e.__class__.__name__, str(e)))
    except (KeyboardInterrupt, EOFError):
        pass


if __name__ == "__main__":
    main()



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from github import Github
from github.AuthenticatedUser import AuthenticatedUser
from msm import MycroftSkillsManager

from msk.lazy import Lazy, unset
from msk.util import ask_for_github_token


class GlobalContext:
    lang = Lazy(unset)  # type: str
    msm = Lazy(unset)  # type: MycroftSkillsManager
    use_token = Lazy(unset)  # type: bool
    branch = Lazy(unset)  # type: str
    github = Lazy(lambda s: ask_for_github_token())  # type: Github
    user = Lazy(lambda s: s.github.get_user())  # type: AuthenticatedUser



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from contextlib import suppress
from git import Git, GitCommandError
from github.Repository import Repository
from msm import SkillRepo, SkillEntry
from os.path import join
from subprocess import call

from msk.exceptions import AlreadyUpdated, NotUploaded
from msk.global_context import GlobalContext
from msk.lazy import Lazy
from msk.util import skill_repo_name


class RepoData(GlobalContext):
    msminfo = Lazy(lambda s: s.msm.repo)  # type: SkillRepo
    git = Lazy(lambda s: Git(s.msminfo.path))  # type: Git
    hub = Lazy(
        lambda s: s.github.get_repo(skill_repo_name(s.msminfo.url))
    )  # type: Repository
    fork = Lazy(
        lambda s: s.github.get_user().create_fork(s.hub)
    )  # type: Repository

    def push_to_fork(self, branch: str):
        remotes = self.git.remote().split("\n")
        command = "set-url" if "fork" in remotes else "add"
        self.git.remote(command, "fork", self.fork.html_url)

        # Use call to ensure the environment variable GIT_ASKPASS is used
        call(
            ["git", "push", "-u", "fork", branch, "--force"],
            cwd=self.msminfo.path,
        )

    def checkout_branch(self, branch):
        with suppress(GitCommandError):
            self.git.branch("-D", branch)
        try:
            self.git.checkout(b=branch)
        except GitCommandError:
            self.git.checkout(branch)


class SkillData(GlobalContext):
    def __init__(self, skill: SkillEntry):
        self.entry = skill

    name = property(lambda self: self.entry.name)
    repo = Lazy(lambda s: RepoData())  # type: RepoData
    repo_git = Lazy(
        lambda s: Git(join(s.repo.msminfo.path, s.submodule_name))
    )  # type: Git
    repo_branch = Lazy(
        lambda s: s.repo_git.symbolic_ref("refs/remotes/origin/HEAD")
    )
    git = Lazy(lambda s: Git(s.entry.path))  # type: Git
    hub = Lazy(
        lambda s: s.github.get_repo(skill_repo_name(s.entry.url))
    )  # type: Repository

    @Lazy
    def submodule_name(self):
        name_to_path = {
            name: path
            for name, path, url, sha in self.repo.msminfo.get_skill_data()
        }
        if self.name not in name_to_path:
            raise NotUploaded(
                f"The skill {self.name} has not yet been uploaded to the "
                "skill store"
            )
        return name_to_path[self.name]

    def upgrade(self) -> str:
        skill_module = self.submodule_name
        submod = Git(join(self.repo.msminfo.path, skill_module))
        submod.remote("set-head", "origin", "-a")
        self.repo.msminfo.update()
        self.repo_git.fetch()
        self.repo_git.reset(self.repo_branch, hard=True)

        upgrade_branch = "upgrade-" + self.name
        self.repo.checkout_branch(upgrade_branch)

        if not self.repo.git.diff(skill_module) and self.repo.git.ls_files(
            skill_module
        ):
            raise AlreadyUpdated(
                f"The latest version of {self.name} is already uploaded to "
                "the skill repo"
            )
        self.repo.git.add(skill_module)
        self.repo.git.commit(message="Upgrade " + self.name)
        return upgrade_branch

    def add_to_repo(self) -> str:
        self.repo.msminfo.update()
        existing_mods = [
            i.split("\t")[1] for i in self.git.ls_tree("HEAD").split("\n")
        ]
        if self.name not in existing_mods:
            self.repo.git.submodule("add", self.entry.url, self.name)

        # Upgrade skill in case it is outdated
        self.repo_git.fetch()
        self.repo_git.reset(self.repo_branch, hard=True)

        branch_name = "add-" + self.name
        self.repo.checkout_branch(branch_name)
        self.repo.git.add(self.name)
        self.repo.git.commit(message="Add " + self.name)
        return branch_name

    def init_existing(self):
        self.repo.git.submodule("update", "--init", self.submodule_name)



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import atexit

import os
from configparser import NoOptionError
from contextlib import contextmanager
from difflib import SequenceMatcher
from functools import wraps
from git.config import GitConfigParser, get_config_path
from github import Github, GithubException
from github.Repository import Repository
from msm import SkillEntry
from os import chmod
from os.path import join, dirname
from tempfile import mkstemp
from typing import Optional
from glob import glob
from pathlib import Path

from msk import __version__
from msk.exceptions import PRModified, MskException, SkillNameTaken

ASKPASS = '''#!/usr/bin/env python3
print(r"""{token}""")
'''

skills_kit_footer = (
    "<sub>Created with [mycroft-skills-kit]({}) v{}</sub>".format(
        "https://github.com/mycroftai/mycroft-skills-kit", __version__
    )
)

tokendir = str(Path.home()) + "/.mycroft/msk/"
tokenfile = tokendir + "GITHUB_TOKEN"


def register_git_injector(token):
    """Generate a script that writes the token to the git command line tool"""
    fd, tmp_path = mkstemp()
    atexit.register(lambda: os.remove(tmp_path))

    with os.fdopen(fd, "w") as f:
        f.write(ASKPASS.format(token=token.replace('"""', r"\"\"\"").strip()))

    chmod(tmp_path, 0o700)
    os.environ["GIT_ASKPASS"] = tmp_path


def ask_for_github_token() -> Github:
    """Ask for GitHub Token if there isnt stored token
    or stored token is invalid"""
    print("")
    token = get_stored_github_token()
    if token and check_token(token):
        github = Github(token)
        register_git_injector(token)
        return github
    else:
        retry = False
        while True:
            if not retry:
                print(
                    "To authenticate with GitHub a Personal Access Token is "
                    "needed."
                )
                print(
                    "    1. Create one at "
                    "https://github.com/settings/tokens/new"
                )
                print("    2. Give the token a name like mycroft-msk")
                print("    3. Select the scopes")
                print("       [X] repo")
                print("    4. Click Generate Token (at bottom of page)")
                print("    5. Copy the generated token")
                print("    6. Paste it in below")
                print("")
                retry = True
            token = input("Personal Access Token: ")
            if check_token(token):
                github = Github(token)
                store_github_token(token)
                register_git_injector(token)
                return github
            else:
                print("")
                print("Token is incorrect.")
                print(
                    "The token may be missing repo scope permissions,"
                )
                print("or the token is invalid.")
                print("Please retry.")
                print("")


def check_token(token):
    """Check if at GitHub Token has 'repo' in the scope"""
    github = Github(token)
    try:
        _ = github.get_user().login
        _ = github.oauth_scopes
        if "repo" in github.oauth_scopes:
            return True
        else:
            return False
    except Exception:
        return False


def get_stored_github_token():
    """Returns stored GitHub token or false if there isnt
    one or the token is invalid"""
    if os.path.isfile(tokenfile):
        with open(tokenfile, "r") as f:
            token = f.readline()
        if not check_token(token):
            os.remove(tokenfile)
        else:
            return token
    else:
        return False


def store_github_token(token):
    """Ask if user will store GitHUb token and if yes store"""
    print("")
    if ask_yes_no(
        "Do you want msk to store the GitHub Personal Access Token? (Y/n)",
        True,
    ):
        if not os.path.exists(tokendir):
            os.makedirs(tokendir)
        with open(tokenfile, "w") as f:
            f.write(token)
            os.chmod(tokenfile, 0o600)
        print("Your GitHub Personal Access Token is stored in " + tokenfile)
        print("")
    else:
        print("Remember to store your token in a safe place.")
        print("")


def skill_repo_name(url: str):
    return "{}/{}".format(
        SkillEntry.extract_author(url), SkillEntry.extract_repo_name(url)
    )


def ask_input(message: str, validator=lambda x: True, on_fail="Invalid entry"):
    while True:
        resp = input(message + " ").strip()
        try:
            if validator(resp):
                return resp
        except Exception:
            pass
        o = on_fail(resp) if callable(on_fail) else on_fail
        if isinstance(o, str):
            print(o)


def ask_choice(
    message: str, choices: list, allow_empty=False, on_empty=None
) -> Optional[str]:
    if not choices:
        if allow_empty:
            print(on_empty)
            return None
        else:
            raise MskException(on_empty or 'Error with "{}"'.format(message))

    print()
    print(message)
    print(
        "\n".join(
            "{}. {}".format(i + 1, choice) for i, choice in enumerate(choices)
        )
    )
    print()

    def find_match(x):
        if not x and allow_empty:
            return ...
        try:
            return choices[int(x) - 1]
        except (ValueError, IndexError):
            pass

        def calc_conf(y):
            return SequenceMatcher(a=x, b=y).ratio()

        best_choice = max(choices, key=calc_conf)
        best_conf = calc_conf(best_choice)
        if best_conf > 0.8:
            return best_choice
        raise ValueError

    resp = find_match(
        ask_input(">", find_match, "Please enter one of the options.")
    )
    return None if resp is ... else resp


def ask_input_lines(message: str, bullet: str = ">") -> list:
    print(message)
    lines = []
    while len(lines) < 1 or lines[-1]:
        lines.append(ask_input(bullet))
    return lines[:-1]


def ask_yes_no(message: str, default: Optional[bool]) -> bool:
    resp = ask_input(
        message, lambda x: (not x and default is not None) or x in "yYnN"
    )
    return {"n": False, "y": True, "": default}[resp.lower()]


def create_or_edit_pr(
    title: str,
    body: str,
    skills_repo: Repository,
    user,
    branch: str,
    repo_branch: str,
):
    base = repo_branch
    head = "{}:{}".format(user.login, branch)
    pulls = list(skills_repo.get_pulls(base=base, head=head))
    if pulls:
        pull = pulls[0]
        if "mycroft-skills-kit" in pull.body:
            pull.edit(title, body)
        else:
            raise PRModified(
                "Not updating description since it was not autogenerated"
            )
        return pull
    else:
        try:
            return skills_repo.create_pull(title, body, base=base, head=head)
        except GithubException as e:
            if e.status == 422:
                raise SkillNameTaken(title) from e
            raise


def to_camel(snake):
    """time_skill -> TimeSkill"""
    return snake.title().replace("_", "")


def to_snake(camel):
    """TimeSkill -> time_skill"""
    if not camel:
        return camel
    return "".join("_" + x if "A" <= x <= "Z" else x for x in camel).lower()[
        camel[0].isupper():
    ]


@contextmanager
def print_error(exception):
    try:
        yield
    except exception as e:
        print("{}: {}".format(exception.__name__, e))


def read_file(*path):
    with open(join(*path)) as f:
        return f.read()


def read_lines(*path):
    with open(join(*path)) as f:
        return [i for i in (i.strip() for i in f.readlines()) if i]


def serialized(func):
    """Write a serializer by yielding each line of output"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        return "\n".join(
            " ".join(parts) if isinstance(parts, tuple) else parts
            for parts in func(*args, **kwargs)
        )

    return wrapper


def get_licenses():
    licenses = glob(join(dirname(__file__), "licenses", "*.txt"))
    licenses.sort()
    return licenses


GIT_IDENTITY_INFO = """=== Git Identity ===
msk uses Git to save skills to Github and when submitting a skill to the
Mycroft Marketplace. To use Git, Git needs to know your Name and
E-mail address. This is important because every Git commit uses the
information to show the responsible party for the submission.
"""


GIT_MANUAL_CHANGE_INFO = """
Thank you. :)

If you need to change this in the future use

    git --config user.name "My Name"

and

    git --config user.email "me@myhost.com"

"""


def ensure_git_user():
    """Prompt for fullname and email if git config is missing it."""
    conf_path = get_config_path("global")
    with GitConfigParser(conf_path, read_only=False) as conf_parser:

        # Make sure a user section exists
        if "user" not in conf_parser.sections():
            conf_parser.add_section("user")

        # Check for missing options using the ConfigParser and insert them
        # if they're missing.
        name, email = (None, None)
        try:
            name = conf_parser.get(section="user", option="name")
        except NoOptionError:
            pass  # Name doesn't exist deal with it later
        try:
            email = conf_parser.get(section="user", option="email")
        except NoOptionError:
            pass  # E-mail doesn't exist, deal with it later

        if not all((name, email)):
            # Some of the needed config is missing
            print(GIT_IDENTITY_INFO)
            if not name:
                name = input("Please enter Full name: ")
                conf_parser.set("user", "name", name)
            if not email:
                email = input("Please enter e-mail address: ")
                conf_parser.set("user", "email", email)

            print(GIT_MANUAL_CHANGE_INFO)



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import atexit

import re
from argparse import ArgumentParser
import requests
from git import Git, GitCommandError
from github import GithubException
from github.Repository import Repository
from os import makedirs
from os.path import join, exists, isdir, basename, splitext
import shutil
from shutil import rmtree
from subprocess import call
from typing import Callable, Optional
from colorama import Style, Fore, init as colorama_init

from msk.console_action import ConsoleAction
from msk.exceptions import GithubRepoExists, UnrelatedGithubHistory
from msk.lazy import Lazy
from msk.util import (
    ask_input,
    to_camel,
    ask_yes_no,
    ask_input_lines,
    print_error,
    get_licenses,
    ask_choice,
)

readme_template = """# \
<img src="https://raw.githack.com/FortAwesome/Font-Awesome/master/svgs/solid/\
{icon}.svg" card_color="{color}" width="50" height="50" \
style="vertical-align:bottom"/> \
{title_name}
{short_description}

## About
{long_description}

## Examples
{examples}
{credits}
## Category
**{category_primary}**
{categories_other}
## Tags
{tags}
"""

credits_template = """## Credits
{author}
"""

init_template = """from mycroft import MycroftSkill, intent_handler


class {class_name}(MycroftSkill):
    def __init__(self):
        MycroftSkill.__init__(self)

    @intent_handler('{intent_name}.intent')
    def handle_{handler_name}(self, message):
{handler_code}


def create_skill():
    return {class_name}()

"""

gitignore_template = """__pycache__/
*.qmlc
settings.json

"""

settingsmeta_template = """
skillMetadata:
  sections:
    - name: Options << Name of section
      fields:
        - name: internal_python_variable_name
          type: text
          label: Setting Friendly Display Name
          value: ""
          placeholder: demo prompt in the input box
    - name: Login << Name of another section
      fields:
        - type: label
          label: A bit of extra info for the user to understand these settings
        - name: username
          type: text
          label: Username
          value: ""
        - name: password
          type: password
          label: Password
          value: ""
"""

manifest_template = """
# This file details all external dependencies required by your skill.
# If your skill does not require any dependencies, please delete this file
# before submitting a pull request.
#
# To use this file, uncomment the lines you need, and fill in the appropriate
# information.
#
# dependencies:
#   # Pip dependencies on PyPI
#   python:
#     - requests
#     - gensim
#
#   # Install packages with the system package manager
#   # This searches for the provided executable and uses the package names
#   system:
#     # For simple packages, this is all that is necessary
#     all: pianobar piano-dev
#
#     # If the package has a certain name on a different platform:
#     pkcon: pianobar libpiano-dev  # For the mycroft platform
#     apt-get: pianobar libpiano-dev  # For Ubuntu/Debian
#
#   # Require certain executables to be in the PATH for the install to succeed
#   exes:
#     - pianobar
#
#   # Require the installation of other skills before installing this skill
#   skill:
#     - my-other-skill
"""


def pretty_license(path):
    pretty = basename(path)
    pretty = splitext(pretty)[0]
    pretty = pretty.replace("-", " ")
    return pretty


class CreateAction(ConsoleAction):
    def __init__(self, args, name: str = None):
        colorama_init()
        if name:
            self.name = name

    @staticmethod
    def register(parser: ArgumentParser):
        pass

    @Lazy
    def name(self) -> str:
        name_to_skill = {skill.name: skill for skill in self.msm.list()}
        while True:
            name = (
                ask_input(
                    'Enter a short unique skill name (ie. "siren alarm" or '
                    '"pizza orderer"):',
                    lambda x: re.match(r"^[a-zA-Z \-]+$", x),
                    "Please use only letter and spaces.",
                )
                .strip(" -")
                .lower()
                .replace(" ", "-")
            )
            skill = name_to_skill.get(
                name, name_to_skill.get("{}-skill".format(name))
            )
            if skill:
                print(
                    "The skill {} {}already exists".format(
                        skill.name,
                        "by {} ".format(skill.author) * bool(skill.author),
                    )
                )
                if ask_yes_no("Remove it? (y/N)", False):
                    rmtree(skill.path)
                else:
                    continue
            class_name = "{}Skill".format(to_camel(name.replace("-", "_")))
            repo_name = "{}-skill".format(name)
            print()
            print("Class name:", class_name)
            print("Repo name:", repo_name)
            print()
            alright = ask_yes_no("Looks good? (Y/n)", True)
            if alright:
                return name

    path = Lazy(lambda s: join(s.msm.skills_dir, s.name + "-skill"))
    git = Lazy(lambda s: Git(s.path))
    short_description = Lazy(
        lambda s: ask_input(
            "Enter a one line description for your skill (ie. Orders fresh "
            "pizzas from the store):\n-",
        ).capitalize()
    )
    author = Lazy(lambda s: ask_input("Enter author:"))
    intent_lines = Lazy(
        lambda s: [
            i.capitalize()
            for i in ask_input_lines(
                "Enter some example phrases to trigger your skill:", "-"
            )
        ]
    )
    dialog_lines = Lazy(
        lambda s: [
            i.capitalize()
            for i in ask_input_lines(
                "Enter what your skill should say to respond:", "-"
            )
        ]
    )
    intent_entities = Lazy(
        lambda s: set(
            re.findall(
                r"(?<={)[a-z_A-Z]*(?=})", "\n".join(i for i in s.intent_lines)
            )
        )
    )
    dialog_entities = Lazy(
        lambda s: set(
            re.findall(r"(?<={)[a-z_A-Z]*(?=})", "\n".join(s.dialog_lines))
        )
    )
    long_description = Lazy(
        lambda s: "\n\n".join(
            ask_input_lines("Enter a long description:", ">")
        )
        .strip()
        .capitalize()
    )
    icon = Lazy(
        lambda s: ask_input(
            "Go to Font Awesome ({blue}fontawesome.com/cheatsheet{reset}) "
            "and choose an icon."
            "\nEnter the name of the icon (default: robot):".format(
                blue=Fore.BLUE + Style.BRIGHT, reset=Style.RESET_ALL
            ),
            validator=lambda x: x == ""
            or requests.get(
                "https://raw.githack.com/FortAwesome/Font-Awesome/"
                "master/svgs/solid/{x}.svg".format(x=x)
            ).ok,
            on_fail="\n\n{red}Error: The name was not found. Make sure you "
                    "spelled the icon name right,"
                    " and try again.{reset}\n".format(
                        red=Fore.RED + Style.BRIGHT, reset=Style.RESET_ALL
                    ),
        )
    )
    color = Lazy(
        lambda s: ask_input(
            "Pick a {yellow}color{reset} for your icon. Find a color that "
            "matches the color scheme at {blue}mycroft.ai/colors{reset}, or "
            "pick a color at: {blue}color-hex.com.{reset}"
            "\nEnter the color hex code including the # "
            "(default: #22A7F0):".format(
                blue=Fore.BLUE + Style.BRIGHT,
                yellow=Fore.YELLOW,
                reset=Style.RESET_ALL,
            ),
            validator=lambda hex_code: hex_code == ""
            or hex_code[0] == "#"
            and len(hex_code) in [4, 7],
            on_fail="\n{red}Check that you entered a correct hex code, and "
                    "try again.{reset}\n".format(
                red=Fore.RED + Style.BRIGHT, reset=Style.RESET_ALL
            ),
        )
    )

    category_options = [
        "Daily",
        "Configuration",
        "Entertainment",
        "Information",
        "IoT",
        "Music & Audio",
        "Media",
        "Productivity",
        "Transport",
    ]
    category_primary = Lazy(lambda s: s.ask_category_primary())
    categories_other = Lazy(
        lambda s: s.ask_categories_other(s.category_primary)
    )

    tags = Lazy(
        lambda s: [
            i.capitalize()
            for i in ask_input_lines(
                "Enter tags to make it easier to search for your skill "
                "(optional):",
                "-",
            )
        ]
    )

    manifest = Lazy(
        lambda s: manifest_template
        if ask_yes_no(
            message="Does this Skill depend on Python Packages (PyPI), "
                    "System Packages (apt-get/others), or other skills?"
                    "\nThis will create a manifest.yml file for you to "
                    "define the dependencies for your Skill."
                    "\nCheck the Mycroft documentation at "
                    "mycroft.ai/to/skill-dependencies to learn more about "
                    "including dependencies, and the manifest.yml file, in "
                    "Skills. (y/N)",
            default=False,
        )
        else None
    )

    readme = Lazy(
        lambda s: readme_template.format(
            title_name=s.name.replace("-", " ").title(),
            short_description=s.short_description,
            long_description=s.long_description,
            examples="".join('* "{}"\n'.format(i) for i in s.intent_lines),
            credits=credits_template.format(author=s.author),
            icon=s.icon or "robot",
            color=s.color.upper() or "#22A7F0",
            category_primary=s.category_primary,
            categories_other="".join(
                "{}\n".format(i) for i in s.categories_other
            ),
            tags="".join("#{}\n".format(i) for i in s.tags),
        )
    )
    init_file = Lazy(
        lambda s: init_template.format(
            class_name=to_camel(s.name.replace("-", "_")),
            handler_name=s.intent_name.replace(".", "_"),
            handler_code="\n".join(
                " " * 8 * bool(i) + i
                for i in [
                    "{ent} = message.data.get('{ent}')".format(ent=entity)
                    for entity in sorted(s.intent_entities)
                ]
                + [
                    "{ent} = ''".format(ent=entity)
                    for entity in sorted(s.dialog_entities - s.intent_entities)
                ]
                + [""] * bool(s.dialog_entities | s.intent_entities)
                + "self.speak_dialog('{intent}'{args})".format(
                    intent=s.intent_name,
                    args=", data={{\n{}\n}}".format(
                        ",\n".join(
                            "    '{ent}': {ent}".format(ent=entity)
                            for entity in s.dialog_entities | s.intent_entities
                        )
                    )
                    * bool(s.dialog_entities | s.intent_entities),
                ).split("\n")
            ),
            intent_name=s.intent_name,
        )
    )

    intent_name = Lazy(lambda s: ".".join(reversed(s.name.split("-"))))

    def add_locale(self):
        makedirs(join(self.path, "locale", self.lang))
        with open(
            join(self.path, "locale", self.lang, self.intent_name + ".intent"),
            "w",
        ) as f:
            f.write("\n".join(self.intent_lines + [""]))
        with open(
            join(self.path, "locale", self.lang, self.intent_name + ".dialog"),
            "w",
        ) as f:
            f.write("\n".join(self.dialog_lines + [""]))

    def ask_category_primary(self):
        """Ask user to select primary category."""
        category = ask_choice(
            "\nCategories define where the skill will display in the "
            "Marketplace. \nEnter the primary category for your skill: ",
            self.category_options,
            allow_empty=False,
        )
        return category

    def ask_categories_other(self, category_primary):
        """Ask user to select aditional categories."""
        categories_other = []
        while True:
            category_options_formatted = []
            for category in self.category_options:
                if (category == category_primary) or (
                    category in categories_other
                ):
                    category = "*" + category + "*"
                category_options_formatted.append(category)
            category = ask_choice(
                "Enter additional categories (optional):",
                category_options_formatted,
                allow_empty=True,
                on_empty=None,
            )
            if (category is not None) and (category[0] != "*"):
                categories_other.append(category)
            if category is None:
                break
        return categories_other

    def license(self):
        """Ask user to select a license for the repo."""
        license_files = get_licenses()
        print(
            "For uploading a skill a license is required.\n"
            "Choose one of the licenses listed below or add one later.\n"
        )
        for num, pth in zip(range(1, 1 + len(license_files)), license_files):
            print("{}: {}".format(num, pretty_license(pth)))
        choice = ask_input("Choose license above or press Enter to skip?")
        if choice.isdigit():
            index = int(choice) - 1
            shutil.copy(license_files[index], join(self.path, "LICENSE.md"))
            print(
                "\nSome of these require that you insert the project name "
                "and/or author's name. Please check the license file and "
                "add the appropriate information.\n"
            )

    def initialize_template(self, files: set = None):
        git = Git(self.path)

        skill_template = [
            ("", lambda: makedirs(self.path)),
            ("locale", self.add_locale),
            ("__init__.py", lambda: self.init_file),
            ("README.md", lambda: self.readme),
            ("LICENSE.md", self.license),
            (".gitignore", lambda: gitignore_template),
            (
                "settingsmeta.yaml",
                lambda: settingsmeta_template.format(
                    capital_desc=self.name.replace("-", " ").capitalize()
                ),
            ),
            ("manifest.yml", lambda: self.manifest),
            (".git", lambda: git.init()),
        ]

        def cleanup():
            rmtree(self.path)

        if not isdir(self.path):
            atexit.register(cleanup)
        for file, handler in skill_template:
            if files and file not in files:
                continue
            if not exists(join(self.path, file)):
                result = handler()
                if isinstance(result, str) and not exists(
                    join(self.path, file)
                ):
                    with open(join(self.path, file), "w") as f:
                        f.write(result)
        atexit.unregister(cleanup)

    def commit_changes(self):
        if self.git.rev_parse("HEAD", with_exceptions=False) == "HEAD":
            self.git.add(".")
            self.git.commit(message="Initial commit")

    def force_push(
        self, get_repo_name: Callable = None
    ) -> Optional[Repository]:
        if ask_yes_no(
            "Are you sure you want to overwrite the remote github repo? "
            "This cannot be undone and you will lose your commit "
            "history! (y/N)",
            False,
        ):
            repo_name = (get_repo_name and get_repo_name()) or (
                self.name + "-skill"
            )
            repo = self.user.get_repo(repo_name)
            self.git.push("origin", "master", force=True)
            print("Force pushed to GitHub repo:", repo.html_url)
            return repo

    def link_github_repo(
        self, get_repo_name: Callable = None
    ) -> Optional[Repository]:
        if "origin" not in Git(self.path).remote().split("\n"):
            if ask_yes_no(
                "Would you like to link an existing GitHub repo to it? (Y/n)",
                True,
            ):
                repo_name = (get_repo_name and get_repo_name()) or (
                    self.name + "-skill"
                )
                repo = self.user.get_repo(repo_name)
                self.git.remote("add", "origin", repo.html_url)
                self.git.fetch()
                try:
                    self.git.pull("origin", "master")
                except GitCommandError as e:
                    if e.status == 128:
                        raise UnrelatedGithubHistory(repo_name) from e
                    raise
                self.git.push("origin", "master", set_upstream=True)
                print("Linked and pushed to GitHub repo:", repo.html_url)
                return repo

    def create_github_repo(
        self, get_repo_name: Callable = None
    ) -> Optional[Repository]:
        if "origin" not in Git(self.path).remote().split("\n"):
            if ask_yes_no(
                "Would you like to create a GitHub repo for it? (Y/n)", True
            ):
                repo_name = (get_repo_name and get_repo_name()) or (
                    self.name + "-skill"
                )
                try:
                    repo = self.user.create_repo(
                        repo_name, self.short_description
                    )
                except GithubException as e:
                    if e.status == 422:
                        raise GithubRepoExists(repo_name) from e
                    raise
                self.git.remote("add", "origin", repo.html_url)
                call(
                    ["git", "push", "-u", "origin", "master"],
                    cwd=self.git.working_dir,
                )
                print("Created GitHub repo:", repo.html_url)
                return repo
        return None

    def perform(self):
        self.initialize_template()
        self.commit_changes()
        with print_error(GithubRepoExists):
            self.create_github_repo()
        print("Created skill at:", self.path)



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABCMeta, abstractmethod
from argparse import ArgumentParser

from msk.global_context import GlobalContext
from msk.lazy import Lazy
from msk.repo_action import RepoData


class ConsoleAction(GlobalContext, metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def register(parser: ArgumentParser):
        pass

    @abstractmethod
    def perform(self):
        pass

    repo = Lazy(lambda s: RepoData())  # type: RepoData



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from argparse import ArgumentParser

from msk.actions.upgrade import UpgradeAction
from msk.actions.upload import UploadAction
from msk.console_action import ConsoleAction
from msk.exceptions import NotUploaded


class SubmitAction(ConsoleAction):
    def __init__(self, args):
        try:
            self.action = UpgradeAction(args)
        except NotUploaded:
            self.action = UploadAction(args)

    @staticmethod
    def register(parser: ArgumentParser):
        parser.add_argument("skill_folder")

    def perform(self):
        self.action.perform()



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from msm import MsmException


class MskException(MsmException):
    pass


class AlreadyUpdated(MskException):
    pass


class GithubRepoExists(MskException):
    pass


class NotUploaded(MskException):
    pass


class PRModified(MskException):
    pass


class SkillNameTaken(GithubRepoExists):
    """
    raised when you try to submit a skill with an already taken unique name
    """

    pass


class UnrelatedGithubHistory(GithubRepoExists):
    """
    raised when you try to update a remote with unrelated commit history
    """

    pass


class NoGitRepository(MskException):
    """
    Raised when a skill cannot be updated because it does not belong to any
    git repo
    """

    pass



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from itertools import chain, count

import json
import re
from argparse import ArgumentParser
from glob import glob
from os import makedirs
from os.path import join, isdir, basename, isfile, splitext
from random import shuffle
from typing import Dict

from msk.console_action import ConsoleAction
from msk.exceptions import MskException
from msk.global_context import GlobalContext
from msk.lazy import Lazy
from msk.util import (
    ask_yes_no,
    ask_input,
    read_file,
    read_lines,
    ask_choice,
    serialized,
)


class TestCreator(GlobalContext):
    def __init__(self, folder):
        self.folder = folder

    init_file = Lazy(lambda s: join(s.folder, "__init__.py"))
    init_content = Lazy(
        lambda s: read_file(s.init_file) if isfile(s.init_file) else ""
    )
    utterance = Lazy(
        lambda s: ask_input("Enter an example query:", lambda x: x)
    )
    dialogs = Lazy(
        lambda s: [
            splitext(basename(i))[0]
            for i in glob(join(s.folder, "dialog", s.lang, "*.dialog"))
        ]
    )
    expected_dialog = Lazy(
        lambda s: ask_choice(
            "Choose expected dialog (leave empty to skip).",
            s.dialogs,
            allow_empty=True,
            on_empty="No dialogs available. Skipping...",
        )
    )

    padatious_creator = Lazy(
        lambda s: PadatiousTestCreator(s.folder)
    )  # type: PadatiousTestCreator
    adapt_creator = Lazy(
        lambda s: AdaptTestCreator(s.folder)
    )  # type: AdaptTestCreator
    intent_choices = Lazy(
        lambda s: list(
            chain(
                s.adapt_creator.intent_recipes,
                s.padatious_creator.intent_names,
            )
        )
    )

    @Lazy
    def intent_name(self):
        return ask_choice(
            "Which intent would you like to test?",
            self.intent_choices,
            on_empty="No existing intents found. Please create some first",
        )


class AdaptTestCreator(TestCreator):
    """
    Extracts Adapt intents from the source code

    Adapt's intents are made up of two components:
     - The "vocab definitions": words associated with a vocab name
     - The "intent recipe": list of vocab keywords that are required
       and optional for an intent
    """

    intent_regex = (
        (
            r"""@(?:\\\n )*intent_handler  (?:\n )*\(  IntentBuilder  \(  ['"][^'"]*['"]  \)((?:  """  # noqa: E501
            r"""\.  (?:optionally|require)  \(  ['"][a-zA-Z_]+['"]  \))*)\) \n"""  # noqa: E501
            r"""(?: \\\n)* def (?:\\\n )*([a-z_]+)"""
        )
        .replace("  ", r"[\s\n]*")
        .replace(" ", r"\s*")
    )

    parts_regex = (
        r"""\.  (require|optionally)  \(  ['"]([a-zA-Z_]+)['"]  \)""".replace(
            "  ", r"[\s\n]*"
        ).replace(" ", "\s*")  # noqa: W605
    )
    intent_recipe = Lazy(lambda s: s.intent_recipes[s.intent_name])

    @Lazy
    def utterance(self):
        while True:
            utterance = ask_input("Enter an example query:", lambda x: x)

            missing_vocabs = [
                i
                for i in self.intent_recipe["require"]
                if not any(
                    j in utterance.lower() for j in self.vocab_defs.get(i, [])
                )
            ]
            if missing_vocabs:
                print(
                    "Missing the following vocab:", ", ".join(missing_vocabs)
                )
                if ask_yes_no("Continue anyways? (y/N)", False):
                    return utterance
            else:
                return utterance

    def extract_recipe(self, recipe_str):
        parts = {"require": [], "optionally": []}
        for part_match in re.finditer(self.parts_regex, recipe_str):
            parts[part_match.group(1)].append(part_match.group(2))
        return parts

    @Lazy
    def intent_recipes(self) -> Dict[str, Dict[str, list]]:
        return {
            match.group(2): self.extract_recipe(match.group(1))
            for match in re.finditer(self.intent_regex, self.init_content)
        }

    @Lazy
    def vocab_defs(self):
        return {
            splitext(basename(content_file))[0]: list(
                chain(
                    *(
                        map(str.strip, i.lower().split("|"))
                        for i in read_lines(content_file)
                    )
                )
            )
            for content_file in glob(
                join(self.folder, "vocab", self.lang, "*.voc")
            )
            + glob(join(self.folder, "locale", self.lang, "*.voc"))
            + glob(join(self.folder, "regex", self.lang, "*.rx"))
            + glob(join(self.folder, "locale", self.lang, "*.rx"))
        }

    @Lazy
    def utterance_data(self):
        utterance_left = self.utterance.lower()
        utterance_data = {}

        for key, start_message in [
            ("require", "Required"),
            ("optionally", "Optional"),
        ]:
            if not self.intent_recipe[key]:
                continue

            print()
            print("===", start_message, "Tags", "===")
            for vocab_name in sorted(self.intent_recipe[key]):
                vocab_value = ask_input(
                    vocab_name + ":",
                    lambda x: not x or x.lower() in utterance_left,
                    "Response must be in the remaining utterance: "
                    + utterance_left,
                )
                if vocab_value:
                    utterance_data[vocab_name] = vocab_value
                    utterance_left = utterance_left.replace(
                        vocab_value.lower(), ""
                    )

        return utterance_data

    @Lazy
    @serialized
    def recipe_str(self):
        for key, name in [("require", "Required"), ("optionally", "Optional")]:
            if not self.intent_recipe[key]:
                continue

            yield ""
            yield "===", name, "Vocab", "==="
            for vocab_name in sorted(self.intent_recipe[key]):
                words = self.vocab_defs.get(vocab_name, ["?"])
                yield "{}: {}".format(
                    vocab_name,
                    ", ".join(words[:6] + ["..."] * (len(words) > 6)),
                )

    @Lazy
    def test_case(self) -> dict:
        if self.intent_name not in self.intent_recipes:
            return {}

        print(self.recipe_str)
        print()

        test_case = {"utterance": self.utterance}
        if self.utterance_data:
            test_case["intent"] = self.utterance_data
        test_case["intent_type"] = self.intent_name
        if self.expected_dialog:
            test_case["expected_dialog"] = self.expected_dialog
        return test_case


class PadatiousTestCreator(TestCreator):
    intent_files = Lazy(
        lambda s: glob(join(s.folder, "vocab", s.lang, "*.intent"))
        + glob(join(s.folder, "locale", s.lang, "*.intent"))
    )
    intent_names = Lazy(
        lambda s: {
            basename(intent_file): intent_file
            for intent_file in s.intent_files
        }
    )
    intent_file = Lazy(lambda s: s.intent_names.get(s.intent_name, ""))
    entities = Lazy(
        lambda s: {
            splitext(basename(entity_file))[0]: read_lines(entity_file)
            for entity_file in glob(
                join(s.folder, "vocab", s.lang, "*.entity")
            )
            + glob(join(s.folder, "locale", s.lang, "*.entity"))
        }
    )

    intent_lines = Lazy(lambda s: read_lines(s.intent_file))
    entity_names = Lazy(
        lambda s: set(
            re.findall(r"(?<={)[a-z_]+(?=})", "\n".join(s.intent_lines))
        )
    )

    @Lazy
    @serialized
    def entities_str(self) -> str:
        if not self.entities:
            return
        yield "=== Entity Examples ==="
        for entity_name, lines in sorted(self.entities.items()):
            sample = ", ".join(lines)
            yield "{}: {}".format(
                entity_name, sample[:50] + "..." * (len(sample) > 50)
            )

    @Lazy
    @serialized
    def intent_str(self) -> str:
        shuffle(self.intent_lines)
        yield "=== Intent Examples ==="
        yield "\n".join(
            self.intent_lines[:6] + ["..."] * (len(self.intent_lines) > 6)
        )

    @Lazy
    def utterance_data(self) -> dict:
        utterance_data = {}
        utterance_left = self.utterance

        print()
        print("=== Entity Tags ===")
        for entity_name in sorted(self.entity_names):
            entity_value = ask_input(
                entity_name + ":",
                lambda x: not x or x in utterance_left,
                "Response must be in the remaining utterance: "
                + utterance_left,
            )
            if entity_value:
                utterance_data[entity_name] = entity_value
                utterance_left = utterance_left.replace(entity_value, "")
        return utterance_data

    @Lazy
    def test_case(self) -> dict:
        if self.intent_name not in self.intent_names:
            return {}

        print()
        print(self.intent_str)
        print()
        if self.entities_str:
            print(self.entities_str)
            print()

        test_case = {"utterance": self.utterance}
        if self.entity_names and self.utterance_data:
            test_case["intent"] = self.utterance_data
        test_case["intent_type"] = self.intent_name
        if self.expected_dialog:
            test_case["expected_dialog"] = self.expected_dialog
        return test_case


class CreateTestAction(ConsoleAction):
    def __init__(self, args):
        self.folder = args.skill_folder

    @staticmethod
    def register(parser: ArgumentParser):
        parser.add_argument("skill_folder")

    def find_intent_test_file(self, intent_name):
        def create_name(i):
            return join(
                self.folder,
                "test",
                "intent",
                "{}.{}.intent.json".format(intent_name, i),
            )

        for i in count():
            name = create_name(i)
            if not isfile(name):
                return name

    def perform(self):
        if not isdir(self.folder):
            raise MskException(
                "Skill folder at {} does not exist".format(self.folder)
            )

        if not isfile(join(self.folder, "__init__.py")):
            if not ask_yes_no(
                "Folder doesn't appear to be a skill. Continue? (y/N)", False
            ):
                return

        makedirs(join(self.folder, "test", "intent"), exist_ok=True)

        creator = TestCreator(self.folder)
        test_case = (
            creator.adapt_creator.test_case
            or creator.padatious_creator.test_case
        )

        intent_test_file = self.find_intent_test_file(creator.intent_name)
        with open(intent_test_file, "w") as f:
            json.dump(test_case, f, indent=4, sort_keys=True)
        print("Generated test file:", intent_test_file)



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Light
# (see https://github.com/MatthewScholefield/mycroft-light).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='msk',
    version='0.3.17rc1',  # Also update in msk/__init__.py
    packages=['msk', 'msk.actions'],
    package_data={'msk': ['licenses/*']},
    install_requires=['GitPython>=3.0.5', 'msm~=0.8.9', 'pygithub',
                      'requests', 'colorama'],
    url='https://github.com/MycroftAI/mycroft-skills-kit',
    license='Apache-2.0',
    author='Mycroft AI',
    author_email='support@mycroft.ai',
    maintainer='Matthew Scholefield',
    maintainer_email='matthew331199@gmail.com',
    description='Mycroft Skills Kit',
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        'console_scripts': {
            'msk=msk.__main__:main'
        }
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Kit
# (see https://github.com/MycroftAI/mycroft-skills-kit).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from argparse import ArgumentParser
from genericpath import samefile
from git import Git
from github.Repository import Repository
from msm import MycroftSkillsManager

from msk.console_action import ConsoleAction
from msk.exceptions import NotUploaded
from msk.repo_action import SkillData
from msk.util import skills_kit_footer, create_or_edit_pr

body_template = (
    """
This upgrades {skill_name} to include the following new commits:

{commits}

"""
    + skills_kit_footer
)


class UpgradeAction(ConsoleAction):
    def __init__(self, args):
        msm = MycroftSkillsManager()
        skill_matches = [
            skill
            for skill in msm.list()
            if skill.is_local and samefile(skill.path, args.skill_folder)
        ]
        if not skill_matches:
            raise NotUploaded(
                "Skill at folder not uploaded to store: "
                f"{args.skill_folder}"
            )

        self.skill = SkillData(skill_matches[0])
        self.skill.init_existing()  # Verifies the skill exists

    @staticmethod
    def register(parser: ArgumentParser):
        pass  # Implemented in SubmitAction

    def create_pr_message(
        self, skill_git: Git, skill_repo: Repository
    ) -> tuple:
        """Create a list of changes for PR content from git commits."""
        title = "Upgrade " + self.skill.name
        body = body_template.format(
            skill_name=self.skill.name,
            commits="\n".join(
                " - [{}]({})".format(
                    skill_git.show("-s", sha, format="%s"),
                    skill_repo.get_commit(sha).html_url,
                )
                for sha in skill_git.rev_list(
                    "--ancestry-path",
                    "{}..{}".format(self.skill.entry.sha, "HEAD"),
                ).split("\n")
            ),
        )
        return title, body

    def perform(self):
        print("Upgrading an existing skill in the skill repo...")
        upgrade_branch = self.skill.upgrade()
        self.repo.push_to_fork(upgrade_branch)
        title, body = self.create_pr_message(self.skill.git, self.skill.hub)
        print()
        print("===", title, "===")
        print(body)
        print()
        pull = create_or_edit_pr(
            title, body, self.repo.hub, self.user, upgrade_branch, self.branch
        )
        print("Created PR at:", pull.html_url)
